
#include <stdlib.h>

#include "core/system.h"
#include "core/gl_util.h"
#include "core/vmath.h"
#include "core/polygon.h"
#include "core/obb.h"
#include "mesh.h"
#include "skeletal_model.h"


void SSBoneFrame_InitSSAnim(struct ss_animation_s *ss_anim, uint32_t anim_type_id);

void SkeletalModel_Clear(skeletal_model_p model)
{
    if(model != NULL)
    {
        if(model->mesh_tree)
        {
            model->mesh_count = 0;
            free(model->mesh_tree);
            model->mesh_tree = NULL;
        }

        if(model->collision_map)
        {
            free(model->collision_map);
            model->collision_map = NULL;
        }

        if(model->animation_count)
        {
            animation_frame_p anim = model->animations;
            for(uint16_t i = 0; i < model->animation_count; i++, anim++)
            {
                if(anim->state_change_count)
                {
                    for(uint16_t j = 0; j < anim->state_change_count; j++)
                    {
                        anim->state_change[j].anim_dispatch_count = 0;
                        free(anim->state_change[j].anim_dispatch);
                        anim->state_change[j].anim_dispatch = NULL;
                        anim->state_change[j].id = 0;
                    }
                    anim->state_change_count = 0;
                    free(anim->state_change);
                    anim->state_change = NULL;
                }

                if(anim->frames_count)
                {
                    for(uint16_t j = 0; j < anim->frames_count; j++)
                    {
                        if(anim->frames[j].bone_tag_count)
                        {
                            anim->frames[j].bone_tag_count = 0;
                            free(anim->frames[j].bone_tags);
                            anim->frames[j].bone_tags = NULL;
                        }
                    }
                    anim->frames_count = 0;
                    anim->max_frame = 0;
                    free(anim->frames);
                    anim->frames = NULL;
                }
                
                while(anim->commands)
                {
                    animation_command_p next_command = anim->commands->next;
                    free(anim->commands);
                    anim->commands = next_command;
                }
                
                while(anim->effects)
                {
                    animation_effect_p next_effect = anim->effects->next;
                    free(anim->effects);
                    anim->effects = next_effect;
                }
            }
            model->animation_count= 0;
            free(model->animations);
            model->animations = NULL;
        }
    }
}


void SkeletalModel_GenParentsIndexes(skeletal_model_p model)
{
    int stack = 0;
    uint16_t parents[model->mesh_count];

    parents[0] = 0;
    model->mesh_tree[0].parent = 0;                                             // root
    for(uint16_t i = 1; i < model->mesh_count; i++)
    {
        model->mesh_tree[i].parent = i - 1;
        if(model->mesh_tree[i].flag & 0x01)                                     // POP
        {
            if(stack > 0)
            {
                model->mesh_tree[i].parent = parents[stack];
                stack--;
            }
        }
        if(model->mesh_tree[i].flag & 0x02)                                     // PUSH
        {
            if(stack + 1 < (int16_t)model->mesh_count)
            {
                stack++;
                parents[stack] = model->mesh_tree[i].parent;
            }
        }
    }
    
    for(uint16_t i = 1; i + 1 < model->mesh_count; i++)
    {
        for(uint16_t j = 1; j + i < model->mesh_count; j++)
        {
            uint16_t m1 = model->collision_map[j];
            uint16_t m2 = model->collision_map[j + 1];
            if(model->mesh_tree[m1].parent > model->mesh_tree[m2].parent)
            {
                uint16_t t = model->collision_map[j];
                model->collision_map[j] = model->collision_map[j + 1];
                model->collision_map[j + 1] = t;
            }
        }
    }
}


void SkeletalModel_FillTransparency(skeletal_model_p model)
{
    model->transparency_flags = MESH_FULL_OPAQUE;
    for(uint16_t i = 0; i < model->mesh_count; i++)
    {
        if(model->mesh_tree[i].mesh_base->transparency_polygons != NULL)
        {
            model->transparency_flags = MESH_HAS_TRANSPARENCY;
            return;
        }
    }
}


void SkeletalModel_CopyMeshes(mesh_tree_tag_p dst, mesh_tree_tag_p src, int tags_count)
{
    for(int i = 0; i < tags_count; i++)
    {
        dst[i].mesh_base = src[i].mesh_base;
    }
}


void BoneFrame_Copy(bone_frame_p dst, bone_frame_p src)
{
    if(dst->bone_tag_count < src->bone_tag_count)
    {
        dst->bone_tags = (bone_tag_p)realloc(dst->bone_tags, src->bone_tag_count * sizeof(bone_tag_t));
    }
    dst->bone_tag_count = src->bone_tag_count;
    vec3_copy(dst->pos, src->pos);
    vec3_copy(dst->centre, src->centre);
    vec3_copy(dst->bb_max, src->bb_max);
    vec3_copy(dst->bb_min, src->bb_min);

    for(uint16_t i = 0; i < dst->bone_tag_count; i++)
    {
        vec4_copy(dst->bone_tags[i].qrotate, src->bone_tags[i].qrotate);
        vec3_copy(dst->bone_tags[i].offset, src->bone_tags[i].offset);
    }
}


void SSBoneFrame_CreateFromModel(ss_bone_frame_p bf, skeletal_model_p model)
{
    vec3_set_zero(bf->bb_min);
    vec3_set_zero(bf->bb_max);
    vec3_set_zero(bf->centre);
    vec3_set_zero(bf->pos);
    bf->transform = NULL;
    bf->bone_tag_count = 0;
    bf->bone_tags = NULL;
    
    SSBoneFrame_InitSSAnim(&bf->animations, ANIM_TYPE_BASE);
    bf->animations.model = model;
    if(model)
    {
        bf->bone_tag_count = model->mesh_count;
        bf->bone_tags = (ss_bone_tag_p)malloc(bf->bone_tag_count * sizeof(ss_bone_tag_t));
        bf->bone_tags[0].parent = NULL;                                         // root
        for(uint16_t i = 0; i < bf->bone_tag_count; i++)
        {
            bf->bone_tags[i].index = i;
            bf->bone_tags[i].is_hidden = 0x00;
            bf->bone_tags[i].mesh_base = model->mesh_tree[i].mesh_base;
            bf->bone_tags[i].mesh_replace = NULL;
            bf->bone_tags[i].mesh_skin = NULL;
            bf->bone_tags[i].mesh_slot = NULL;
            bf->bone_tags[i].skin_map = NULL;
            bf->bone_tags[i].alt_anim = NULL;
            bf->bone_tags[i].body_part = model->mesh_tree[i].body_part;

            vec3_copy(bf->bone_tags[i].offset, model->mesh_tree[i].offset);
            vec4_set_zero(bf->bone_tags[i].qrotate);
            Mat4_E_macro(bf->bone_tags[i].transform);
            Mat4_E_macro(bf->bone_tags[i].full_transform);

            if(i > 0)
            {
                bf->bone_tags[i].parent = bf->bone_tags + model->mesh_tree[i].parent;
            }
        }
    }
}


void SSBoneFrame_InitSSAnim(struct ss_animation_s *ss_anim, uint32_t anim_type_id)
{
    ss_anim->anim_ext_flags = 0x00;
    ss_anim->anim_frame_flags = 0x00;
    ss_anim->type = anim_type_id;
    ss_anim->enabled = 1;
    ss_anim->model = NULL;
    ss_anim->onFrame = NULL;
    ss_anim->onEndFrame = NULL;
    ss_anim->targeting_bone = 0x00;
    ss_anim->targeting_flags = 0x0000;
    vec3_set_zero(ss_anim->target);
    vec4_set_zero_angle(ss_anim->current_mod);
    ss_anim->bone_direction[0] = 0.0f;
    ss_anim->bone_direction[1] = 1.0f;
    ss_anim->bone_direction[2] = 0.0f;
    ss_anim->targeting_limit[0] = 0.0f;
    ss_anim->targeting_limit[1] = 1.0f;
    ss_anim->targeting_limit[2] = 0.0f;
    ss_anim->targeting_limit[3] =-1.0f;
    vec3_set_one(ss_anim->targeting_axis_mod);

    ss_anim->frame_time = 0.0f;
    ss_anim->next_state = 0;
    ss_anim->next_state_heavy = -1;
    ss_anim->lerp = 0.0;
    ss_anim->current_animation = 0;
    ss_anim->current_frame = 0;
    ss_anim->next_animation = 0;
    ss_anim->next_frame = 0;
    ss_anim->period = 1.0f / 30.0f;

    ss_anim->next = NULL;
    ss_anim->prev = NULL;
}


void SSBoneFrame_Clear(ss_bone_frame_p bf)
{
    if(bf && bf->bone_tag_count)
    {
        for(uint16_t i = 0; i < bf->bone_tag_count; i++)
        {
            if(bf->bone_tags[i].skin_map)
            {
                free(bf->bone_tags[i].skin_map);
            }
        }
        
        free(bf->bone_tags);
        bf->bone_tag_count = 0;
        bf->bone_tags = NULL;
    }

    for(ss_animation_p ss_anim = bf->animations.next; ss_anim;)
    {
        ss_animation_p ss_anim_next = ss_anim->next;
        ss_anim->next = NULL;
        free(ss_anim);
        ss_anim = ss_anim_next;
    }
    bf->animations.next = NULL;
}


void SSBoneFrame_Copy(struct ss_bone_frame_s *dst, struct ss_bone_frame_s *src)
{
    if(dst->bone_tag_count == src->bone_tag_count)
    {
        vec3_copy(dst->pos, src->pos);
        vec3_copy(dst->centre, src->centre);
        vec3_copy(dst->bb_max, src->bb_max);
        vec3_copy(dst->bb_min, src->bb_min);
        
        dst->animations.model = src->animations.model;
        dst->animations.enabled = src->animations.enabled;
        dst->animations.current_animation = src->animations.current_animation;
        dst->animations.current_frame = src->animations.current_frame;
        dst->animations.next_animation = src->animations.next_animation;
        dst->animations.next_frame = src->animations.next_frame;
        dst->animations.lerp = src->animations.lerp;
        dst->animations.frame_time = src->animations.frame_time;
        dst->animations.next_state = src->animations.next_state;
    }
}


void SSBoneFrame_Update(struct ss_bone_frame_s *bf, float time)
{
    float t = 1.0f - bf->animations.lerp;
    ss_bone_tag_p btag = bf->bone_tags;
    bone_tag_p src_btag, next_btag;
    skeletal_model_p model = bf->animations.model;
    animation_frame_p curr_anim = model->animations + bf->animations.current_animation;
    animation_frame_p next_anim = model->animations + bf->animations.next_animation;
    bone_frame_p curr_bf = curr_anim->frames + bf->animations.current_frame;
    bone_frame_p next_bf = next_anim->frames + bf->animations.next_frame;
    
    vec3_interpolate_macro(bf->bb_max, curr_bf->bb_max, next_bf->bb_max, bf->animations.lerp, t);
    vec3_interpolate_macro(bf->bb_min, curr_bf->bb_min, next_bf->bb_min, bf->animations.lerp, t);
    vec3_interpolate_macro(bf->centre, curr_bf->centre, next_bf->centre, bf->animations.lerp, t);
    vec3_interpolate_macro(bf->pos, curr_bf->pos, next_bf->pos, bf->animations.lerp, t);
    
    next_btag = next_bf->bone_tags;
    src_btag = curr_bf->bone_tags;
    for(uint16_t k = 0; k < curr_bf->bone_tag_count; k++, btag++, src_btag++, next_btag++)
    {
        vec3_interpolate_macro(btag->offset, src_btag->offset, next_btag->offset, bf->animations.lerp, t);
        vec3_copy(btag->transform + 12, btag->offset);
        btag->transform[15] = 1.0f;
        if(k == 0)
        {
            vec3_add(btag->transform + 12, btag->transform + 12, bf->pos);
            vec4_slerp(btag->qrotate, src_btag->qrotate, next_btag->qrotate, bf->animations.lerp);
        }
        else
        {
            bone_tag_p ov_src_btag = src_btag;
            bone_tag_p ov_next_btag = next_btag;
            float ov_lerp = bf->animations.lerp;
            if(btag->alt_anim && btag->alt_anim->model && btag->alt_anim->enabled && (btag->alt_anim->model->mesh_tree[k].replace_anim != 0))
            {
                curr_anim = btag->alt_anim->model->animations + btag->alt_anim->current_animation;
                next_anim = btag->alt_anim->model->animations + btag->alt_anim->next_animation;
                bone_frame_p ov_curr_bf = curr_anim->frames + btag->alt_anim->current_frame;
                bone_frame_p ov_next_bf = next_anim->frames + btag->alt_anim->next_frame;
                ov_lerp = btag->alt_anim->lerp;
                ov_src_btag = ov_curr_bf->bone_tags + k;
                ov_next_btag = ov_next_bf->bone_tags + k;
            }
            vec4_slerp(btag->qrotate, ov_src_btag->qrotate, ov_next_btag->qrotate, ov_lerp);
        }
        Mat4_set_qrotation(btag->transform, btag->qrotate);
    }

    /*
     * build absolute coordinate matrix system
     */
    btag = bf->bone_tags;
    Mat4_Copy(btag->full_transform, btag->transform);
    Mat4_Copy(btag->orig_transform, btag->transform);
    btag++;
    for(uint16_t k = 1; k < curr_bf->bone_tag_count; k++, btag++)
    {
        Mat4_Mat4_mul(btag->full_transform, btag->parent->full_transform, btag->transform);
        Mat4_Copy(btag->orig_transform, btag->full_transform);
    }

    for(ss_animation_p ss_anim = &bf->animations; ss_anim; ss_anim = ss_anim->next)
    {
        SSBoneFrame_TargetBoneToSlerp(bf, ss_anim, time);
    }
}


void SSBoneFrame_RotateBone(struct ss_bone_frame_s *bf, const float q_rotate[4], int bone)
{
    float tr[16], q[4];
    ss_bone_tag_p b_tag = b_tag = bf->bone_tags + bone;
    
    vec4_copy(q, q_rotate);
    Mat4_E(tr);
    Mat4_RotateQuaternion(tr, q);
    vec4_copy(q, b_tag->transform + 12);
    Mat4_Mat4_mul(b_tag->transform, tr, b_tag->transform);
    vec4_copy(b_tag->transform + 12, q);
    for(uint16_t i = bone; i < bf->bone_tag_count; i++)
    {
        ss_bone_tag_p btag = bf->bone_tags + i;
        if(btag->parent)
        {
            Mat4_Mat4_mul(btag->full_transform, btag->parent->full_transform, btag->transform);
        }
        else
        {
            Mat4_Copy(btag->full_transform, btag->transform);
        }
    }
}


int  SSBoneFrame_CheckTargetBoneLimit(struct ss_bone_frame_s *bf, struct ss_animation_s *ss_anim)
{
    ss_bone_tag_p b_tag = b_tag = bf->bone_tags + ss_anim->targeting_bone;
    float target_dir[3], target_local[3], limit_dir[3], t;

    Mat4_vec3_mul_inv(target_local, bf->transform, ss_anim->target);
    if(b_tag->parent)
    {
        Mat4_vec3_mul_inv(target_local, b_tag->parent->full_transform, target_local);
    }
    vec3_sub(target_dir, target_local, b_tag->transform + 12);
    vec3_norm(target_dir, t);
    vec3_copy(limit_dir, ss_anim->targeting_limit);

    if((ss_anim->targeting_limit[3] == -1.0f) ||
       (vec3_dot(limit_dir, target_dir) > ss_anim->targeting_limit[3]))
    {
        return 1;
    }

    return 0;
}


void SSBoneFrame_TargetBoneToSlerp(struct ss_bone_frame_s *bf, struct ss_animation_s *ss_anim, float time)
{
    if(ss_anim->anim_ext_flags & ANIM_EXT_TARGET_TO)
    {
        ss_bone_tag_p b_tag = bf->bone_tags + ss_anim->targeting_bone;
        float clamped_q[4], q[4], target_dir[3], target_local[3], bone_dir[3];

        Mat4_vec3_mul_inv(target_local, bf->transform, ss_anim->target);
        if(b_tag->parent)
        {
            Mat4_vec3_mul_inv(target_local, b_tag->parent->full_transform, target_local);
        }
        vec3_sub(target_dir, target_local, b_tag->transform + 12);
        if(ss_anim->targeting_flags & ANIM_TARGET_OWERRIDE_ANIM)
        {
            Mat4_vec3_rot_macro(bone_dir, b_tag->transform, ss_anim->bone_direction);
        }
        else
        {
            vec3_copy(bone_dir, ss_anim->bone_direction);
        }

        vec4_GetQuaternionRotation(q, bone_dir, target_dir);
        if(ss_anim->targeting_flags & ANIM_TARGET_USE_AXIS_MOD)
        {
            q[0] *= ss_anim->targeting_axis_mod[0];
            q[1] *= ss_anim->targeting_axis_mod[1];
            q[2] *= ss_anim->targeting_axis_mod[2];
            q[3] = 1.0f - vec3_sqabs(q);
            q[3] = sqrtf(q[3]);
        }
        if(q[3] < ss_anim->targeting_limit[3])
        {
            vec4_clampw(q, ss_anim->targeting_limit[3]);
        }
        vec4_slerp_to(clamped_q, ss_anim->current_mod, q, time * M_PI / 1.3f);
        vec4_copy(ss_anim->current_mod, clamped_q);
        SSBoneFrame_RotateBone(bf, ss_anim->current_mod, ss_anim->targeting_bone);
    }
    else if(ss_anim->current_mod[3] < 1.0f)
    {       
        if(ss_anim->current_mod[3] < 0.99f)
        {
            float zero_ang[4] = {0.0f, 0.0f, 0.0f, 1.0f};
            float clamped_q[4];
            vec4_slerp_to(clamped_q, ss_anim->current_mod, zero_ang, time * M_PI / 1.3f);
            vec4_copy(ss_anim->current_mod, clamped_q);
            SSBoneFrame_RotateBone(bf, ss_anim->current_mod, ss_anim->targeting_bone);
        }
        else
        {
            vec4_set_zero_angle(ss_anim->current_mod);
        }
    }
}


void SSBoneFrame_SetTarget(struct ss_animation_s *ss_anim, uint16_t targeted_bone, const float target_pos[3], const float bone_dir[3])
{
    ss_anim->targeting_bone = targeted_bone;
    vec3_copy(ss_anim->target, target_pos);
    vec3_copy(ss_anim->bone_direction, bone_dir);
}


void SSBoneFrame_SetTargetingAxisMod(struct ss_animation_s *ss_anim, const float mod[3])
{
    if(mod)
    {
        vec3_copy(ss_anim->targeting_axis_mod, mod);
        ss_anim->targeting_flags |= ANIM_TARGET_USE_AXIS_MOD;
    }
    else
    {
        vec3_set_one(ss_anim->targeting_axis_mod);
        ss_anim->targeting_flags &= ~ANIM_TARGET_USE_AXIS_MOD;
    }
}


void SSBoneFrame_SetTargetingLimit(struct ss_animation_s *ss_anim, const float limit[4])
{
    if(limit)
    {
        vec4_copy(ss_anim->targeting_limit, limit);
    }
    else
    {
        ss_anim->targeting_limit[3] = -1.0f;
    }
}


struct ss_animation_s *SSBoneFrame_AddOverrideAnim(struct ss_bone_frame_s *bf, struct skeletal_model_s *sm, uint16_t anim_type_id)
{
    if(!sm || (sm->mesh_count == bf->bone_tag_count))
    {
        ss_animation_p ss_anim = (ss_animation_p)malloc(sizeof(ss_animation_t));
        SSBoneFrame_InitSSAnim(ss_anim, anim_type_id);
        ss_anim->model = sm;

        ss_anim->next = bf->animations.next;
        if(bf->animations.next)
        {
            bf->animations.next->prev = ss_anim;
        }
        bf->animations.next = ss_anim;
        
        return ss_anim;
    }

    return NULL;
}


struct ss_animation_s *SSBoneFrame_GetOverrideAnim(struct ss_bone_frame_s *bf, uint16_t anim_type)
{
    for(ss_animation_p p = &bf->animations; p; p = p->next)
    {
        if(p->type == anim_type)
        {
            return p;
        }
    }
    return NULL;
}


void SSBoneFrame_EnableOverrideAnimByType(struct ss_bone_frame_s *bf, uint16_t anim_type)
{
    for(ss_animation_p ss_anim = &bf->animations; ss_anim; ss_anim = ss_anim->next)
    {
        if(ss_anim->type == anim_type)
        {
            ss_anim->enabled = 1;
            for(uint16_t i = 0; i < bf->bone_tag_count; i++)
            {
                mesh_tree_tag_p mtag = ss_anim->model->mesh_tree + i;
                if(mtag->replace_anim != 0)
                {
                    bf->bone_tags[i].alt_anim = ss_anim;
                }
            }
            break;
        }
    }
}


void SSBoneFrame_EnableOverrideAnim(struct ss_bone_frame_s *bf, struct ss_animation_s *ss_anim)
{
    ss_anim->enabled = 1;
    for(uint16_t i = 0; i < bf->bone_tag_count; i++)
    {
        mesh_tree_tag_p mtag = ss_anim->model->mesh_tree + i;
        if(mtag->replace_anim != 0)
        {
            bf->bone_tags[i].alt_anim = ss_anim;
        }
    }
}


void SSBoneFrame_DisableOverrideAnim(struct ss_bone_frame_s *bf, uint16_t anim_type)
{
    for(ss_animation_p ss_anim = &bf->animations; ss_anim; ss_anim = ss_anim->next)
    {
        if(ss_anim->type == anim_type)
        {
            ss_anim->enabled = 0;
            break;
        }
    }
    
    for(uint16_t i = 0; i < bf->bone_tag_count; i++)
    {
        if(bf->bone_tags[i].alt_anim && bf->bone_tags[i].alt_anim->type == anim_type)
        {
            bf->bone_tags[i].alt_anim = NULL;
        }
    }
}


/*
 *******************************************************************************
 */
void Anim_AddCommand(struct animation_frame_s *anim, const animation_command_p command)
{
    animation_command_p *ptr = &anim->commands;
    for(; *ptr; ptr = &((*ptr)->next));
    *ptr = (animation_command_p)malloc(sizeof(animation_command_t));
    **ptr = *command;
    (*ptr)->next = NULL;
}


void Anim_AddEffect(struct animation_frame_s *anim, const animation_effect_p effect)
{
    animation_effect_p *ptr = &anim->effects;
    for(; *ptr; ptr = &((*ptr)->next));
    *ptr = (animation_effect_p)malloc(sizeof(animation_effect_t));
    **ptr = *effect;
    (*ptr)->next = NULL;
}


struct state_change_s *Anim_FindStateChangeByAnim(struct animation_frame_s *anim, int state_change_anim)
{
    if(state_change_anim >= 0)
    {
        state_change_p ret = anim->state_change;
        for(uint16_t i = 0; i < anim->state_change_count; i++, ret++)
        {
            for(uint16_t j = 0; j < ret->anim_dispatch_count; j++)
            {
                if(ret->anim_dispatch[j].next_anim == state_change_anim)
                {
                    return ret;
                }
            }
        }
    }

    return NULL;
}


struct state_change_s *Anim_FindStateChangeByID(struct animation_frame_s *anim, uint32_t id)
{
    state_change_p ret = anim->state_change;
    for(uint16_t i = 0; i < anim->state_change_count; i++, ret++)
    {
        if(ret->id == id)
        {
            return ret;
        }
    }

    return NULL;
}


int Anim_GetAnimDispatchCase(struct ss_animation_s *ss_anim, uint32_t id)
{
    animation_frame_p anim = ss_anim->model->animations + ss_anim->current_animation;
    state_change_p stc = anim->state_change;

    for(uint16_t i = 0; i < anim->state_change_count; i++, stc++)
    {
        if(stc->id == id)
        {
            anim_dispatch_p disp = stc->anim_dispatch;
            for(uint16_t j = 0; j < stc->anim_dispatch_count; j++, disp++)
            {
                if((disp->frame_high >= disp->frame_low) && (ss_anim->current_frame >= disp->frame_low) && (ss_anim->current_frame <= disp->frame_high))
                {
                    return (int)j;
                }
            }
        }
    }

    return -1;
}


void Anim_SetAnimation(struct ss_animation_s *ss_anim, int animation, int frame)
{
    if(ss_anim && ss_anim->model && (animation < ss_anim->model->animation_count))
    {
        animation_frame_p anim = &ss_anim->model->animations[animation];
        ss_anim->lerp = 0.0;
        frame %= anim->max_frame;
        frame = (frame >= 0) ? (frame) : (anim->max_frame - 1 + frame);
        ss_anim->period = 1.0f / 30.0f;

        ss_anim->frame_changing_state = 0x04;

        ss_anim->next_state = anim->state_id;

        ss_anim->next_animation = animation;
        ss_anim->next_frame = frame;
        ss_anim->current_animation = animation;
        ss_anim->current_frame = frame;

        ss_anim->frame_time = (float)frame * ss_anim->period;
    }
}

/*
 * Next frame and next anim calculation function.
 */
int  Anim_SetNextFrame(struct ss_animation_s *ss_anim, float time)
{
    float dt;
    int32_t new_frame;
    animation_frame_p next_anim = ss_anim->model->animations + ss_anim->next_animation;
    state_change_p stc = Anim_FindStateChangeByID(next_anim, ss_anim->next_state);
    
    if((next_anim->state_id == ss_anim->next_state_heavy) && (next_anim->next_anim->state_id == ss_anim->next_state_heavy))
    {
        ss_anim->next_state_heavy = -1;
    }
    
    if(ss_anim->next_state_heavy >= 0)
    {
        state_change_p stc_heavy = Anim_FindStateChangeByID(next_anim, ss_anim->next_state_heavy);
        stc = (stc_heavy) ? (stc_heavy) : (stc);
    }
    
    ss_anim->frame_time = (ss_anim->frame_time >= 0.0f) ? (ss_anim->frame_time) : (0.0f);
    ss_anim->frame_time += time;
    new_frame = ss_anim->frame_time / ss_anim->period;
    dt = ss_anim->frame_time - (float)new_frame * ss_anim->period;
    ss_anim->lerp = dt / ss_anim->period;
    
    ss_anim->frame_changing_state = 0x00;
    
    /*
     * Flag has a highest priority
     */
    if(ss_anim->anim_frame_flags & ANIM_FRAME_LOCK)
    {
        ss_anim->current_frame = 0;
        ss_anim->current_frame = ss_anim->next_frame;
        ss_anim->current_animation = ss_anim->next_animation;
        ss_anim->lerp = 0.0f;
        ss_anim->frame_time = 0.0f;
        return 0x00;
    }
    else if((new_frame + 1 >= next_anim->max_frame) && (ss_anim->anim_frame_flags == ANIM_LOOP_LAST_FRAME))
    {
        ss_anim->next_frame = next_anim->max_frame - 1;
        ss_anim->current_frame = ss_anim->next_frame;
        ss_anim->current_animation = ss_anim->next_animation;
        ss_anim->lerp = 0.0f;
        ss_anim->frame_time = (float)ss_anim->next_frame * ss_anim->period;
        return 0x00;
    }
    
    /*
     * State change check
     */
    if(stc)
    {
        anim_dispatch_p disp = stc->anim_dispatch;
        for(uint16_t i = 0; i < stc->anim_dispatch_count; i++, disp++)
        {
            if((next_anim->max_frame == 1) || 
               (new_frame >= disp->frame_low) && (new_frame <= disp->frame_high) || 
               (ss_anim->next_frame <= disp->frame_high) && (new_frame >= disp->frame_high))
            {
                ss_anim->current_animation = ss_anim->next_animation;
                ss_anim->current_frame = ss_anim->next_frame;
                ss_anim->next_animation = disp->next_anim;
                ss_anim->next_frame = disp->next_frame;
                ss_anim->frame_time = (float)ss_anim->next_frame * ss_anim->period + dt;
                ss_anim->next_state = ss_anim->model->animations[ss_anim->next_animation].state_id;
                ss_anim->frame_changing_state = 0x03;
                return 0x03;
            }
        }
    }
    
    /*
     * Check next anim if frame >= max_frame
     */
    if(new_frame >= next_anim->max_frame)
    {
        ss_anim->current_animation = ss_anim->next_animation;
        ss_anim->current_frame = ss_anim->next_frame;
        ss_anim->next_frame = next_anim->next_frame;
        ss_anim->next_animation = next_anim->next_anim->id;
        ss_anim->frame_time = (float)ss_anim->next_frame * ss_anim->period + dt;
        ss_anim->next_state = ss_anim->model->animations[ss_anim->next_animation].state_id;
        ss_anim->frame_changing_state = 0x02;
        return 0x02;
    }
    
    if(ss_anim->next_frame != new_frame)
    {
        ss_anim->current_animation = ss_anim->next_animation;
        ss_anim->current_frame = ss_anim->next_frame;
        ss_anim->next_frame = new_frame;
        ss_anim->frame_changing_state = 0x01;
        return 0x01;
    }
    
    return 0x00;
}


void SSBoneFrame_FillSkinnedMeshMap(ss_bone_frame_p bf)
{
    uint32_t *ch;
    uint32_t founded_index = 0xFFFFFFFF;
    float tv[3];
    vertex_p v, founded_vertex;
    base_mesh_p mesh_base, mesh_skin;
    ss_bone_tag_p tree_tag = bf->bone_tags;

    for(uint16_t i = 0; i < bf->bone_tag_count; i++, tree_tag++)
    {
        if(!tree_tag->mesh_skin)
        {
            continue;
        }
        if(tree_tag->skin_map)
        {
            free(tree_tag->skin_map);
            tree_tag->skin_map = NULL;
        }
        mesh_base = tree_tag->mesh_base;
        mesh_skin = tree_tag->mesh_skin;
        ch = tree_tag->skin_map = (uint32_t*)malloc(mesh_skin->vertex_count * sizeof(uint32_t));
        v = mesh_skin->vertices;
        for(uint32_t k = 0; k < mesh_skin->vertex_count; k++, v++, ch++)
        {
            *ch = 0xFFFFFFFF;
            founded_index = BaseMesh_FindVertexIndex(mesh_base, v->position);
            if(founded_index != 0xFFFFFFFF)
            {
                founded_vertex = mesh_base->vertices + founded_index;
                vec3_copy(v->position, founded_vertex->position);
                vec3_copy(v->normal, founded_vertex->normal);
            }
            else if(tree_tag->parent)
            {
                vec3_add(tv, v->position, tree_tag->offset);
                founded_index = BaseMesh_FindVertexIndex(tree_tag->parent->mesh_base, tv);
                if(founded_index != 0xFFFFFFFF)
                {
                    founded_vertex = tree_tag->parent->mesh_base->vertices + founded_index;
                    *ch = founded_index;
                    vec3_sub(v->position, founded_vertex->position, tree_tag->offset);
                    vec3_copy(v->normal, founded_vertex->normal);
                }
            }
        }
    }
}